
class RevisionIgnorer(object):
    def __init__(self, lump_builder):
        self.lump_builder = lump_builder

    def process_revision(self, rev, aux_data):
        assert aux_data is None
        builder = self.lump_builder
        builder.revision_header(rev)




class SvnLump:

    def __init__(self):
        self.headers = { }
        self.header_order = [ ]
        self.properties = { }
        self.content = None

    def set_header(self, key, value):
        if not self.headers.has_key(key):
            self.header_order.append(key)
        self.headers[key] = value

    def get_header(self, key):
        return self.headers[key]

    def has_header(self, key):
        return self.headers.has_key(key)

    def delete_header(self, key):
        del self.headers[key]
        self.header_order.remove(key)

    def get_header_keys(self):
        return self.header_order




#
# Purpose:
#
# When we process the output of "svnadmin dump" we want
# to 'stream' some parts ('content') of it via limited
# size buffer to the output. Other content we just want
# to discard.
#
# Regardless if we stream or discard: The file handle of
# the input file handle must be at the right position just
# after the content once we are done with it.
#
# That logic is encapsuled here. Once we identified a
# piece of 'content' we create a ContentTin with it.
#

class NonOut(object):
    def write(self, what):
        pass

class ContentTin:

    chunk_size = 1024**2

    def __init__(self, fh, size, md5sum):
        self.fh = fh
        self.size = size
        self.remaining_bytes = size
        self.md5sum = md5sum
        self.used = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        if not exc_type and not self.used:
            self.empty_to()
        return False
   
    def empty_to(self, out_fh = NonOut()):
        if self.used:
            raise Exception("ContentTins can only be used once")
        while self.remaining_bytes > 0:
            chunk = min(self.chunk_size, self.remaining_bytes)
            input_chunk = self.fh.read(chunk)
            if input_chunk == "":
                raise Exception("read beyond end of file")
            self.remaining_bytes -= len(input_chunk)
            out_fh.write(input_chunk)
        self.used = True

    def discard(self):
        if not self.used:
            self.empty_to()




from string import join

class ParentDirectoryLumpGenerator(object):

    def __init__(self, interesting_paths, lump_builder):
        self.interesting_paths = interesting_paths
        self.lump_builder = lump_builder

    def write_lumps(self):
        paths_to_create_by_length = self._claculate_paths_to_create_by_length()
        for paths_to_create in paths_to_create_by_length:
            for path in paths_to_create:
                self.lump_builder.mkdir(path)

    def _claculate_paths_to_create_by_length(self):
        hashed_paths_by_length = [ None ]
        all_interesting_paths = self.interesting_paths.get_interesting_sub_directories('')
        for path in all_interesting_paths:
            parent_path_elements = path.split('/')[:-1]
            for prefix_length in range(1, len(parent_path_elements)+1):
                prefix = join(parent_path_elements[:prefix_length], '/')
                if prefix_length == len(hashed_paths_by_length):
                    hashed_paths_by_length.append( { prefix: True } )
                else:
                    hashed_paths_by_length[prefix_length][prefix] = True

        paths_by_prefix_length = [ [] ]
        for hashed_paths in hashed_paths_by_length[1:]:
            paths_by_prefix_length.append(sorted(hashed_paths.keys())) # Sorted for easier testing
        return paths_by_prefix_length




from logging import debug

class RevisionMapper(object):

    def __init__(self, config):
        self.start_rev = 0
        if config.start_rev:
            self.start_rev = config.start_rev
        self.revision_map = { }
        self.next_free_rev = 0

    def map_input_rev_to_output_rev(self, input_rev, output_rev):
        if input_rev < self.start_rev:
            raise Exception(
                "Can not map a revsion (%d) below the start-rev (%d)!"
                % ( input_rev, self.start_rev )
            )
        if self.revision_map.has_key(input_rev):
            raise Exception(
                "Revision %d is already mapped!" % ( input_rev )
            )
        
        self.revision_map[input_rev] = output_rev
        debug("    Mapped input revsion input revsion %d to output revsion %d." % ( input_rev, output_rev ) )

    def get_output_rev_for_input_rev(self, input_rev):
        if input_rev < self.start_rev:
            return self.revision_map[self.start_rev]
        else:
            return self.revision_map[input_rev]




from Main import run
#
# The Big Plan
#

# Decide what to do with a revision
from DumpController import DumpController, STRATEGY_DUMP_HEADER, STRATEGY_IGNORE, STRATEGY_SYNTHETIC_DELETES, STRATEGY_DUMP_SCAN, STRATEGY_BOOTSTRAP, DUMP_HEADER_PSEUDO_REV

# Generate the lumps for that
from DumpHeaderGenerator import DumpHeaderGenerator
from BootsTrapper import BootsTrapper
from DumpFilter import DumpFilter, UnsupportedDumpVersionException
from SyntheticDeleter import SyntheticDeleter
from RevisionIgnorer import RevisionIgnorer

# Build the lumps
from LumpBuilder import LumpBuilder
# Fix length fields, drop empty revisions
from LumpPostProcessor import LumpPostProcessor

#
# SVN Abstraction layer
#
from SvnLump import SvnLump
from SvnDumpReader import SvnDumpReader
from SvnDumpWriter import SvnDumpWriter
from SvnRepository import SvnRepository

#
# Helpers
#
from Config import Config
from InterestingPaths import InterestingPaths
from RevisionMapper import RevisionMapper
from CheckedCommandFileHandle import CheckedCommandFileHandle
from ContentTin import ContentTin
from ParentDirectoryLumpGenerator import ParentDirectoryLumpGenerator




from logging import info

class LumpPostProcessor(object):
    def __init__(self, config, revision_mapper, writer):
        self.config = config
        self.revision_mapper = revision_mapper
        self.writer = writer
        self.delayed_revision_header = None
        self.parent_dir_lumps_must_be_injected = True
        self.parent_directory_lump_generator = None
        self.last_revision_passed = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        self._discard_deplayed_revision_header()
        return False

    def write_lump(self, lump):
        if lump.has_header('Revision-number'):
            self._discard_deplayed_revision_header()
            self.delayed_revision_header = lump
            if self.config.keep_empty_revs:
                self._flush_delayed_revision_header()
            if self.parent_dir_lumps_must_be_injected:
                if self.config.create_parent_dirs:
                    self.parent_directory_lump_generator.write_lumps()
                self.parent_dir_lumps_must_be_injected = False
        else:
            self._flush_delayed_revision_header()
            self._process_lump(lump)

    def _discard_deplayed_revision_header(self):
        if not self.delayed_revision_header:
            return
        info('Dropping empty revision.');
        if self.last_revision_passed:
            mapper = self.revision_mapper
            input_rev = int(self.delayed_revision_header.get_header('Revision-number'))
            output_rev = self.last_revision_passed
            mapper.map_input_rev_to_output_rev(input_rev, output_rev)
        self.delayed_revision_header = None

    def _flush_delayed_revision_header(self):
        if not self.delayed_revision_header:
            return
        mapper = self.revision_mapper
        rev = int(self.delayed_revision_header.get_header('Revision-number'))
        mapper.map_input_rev_to_output_rev(rev, rev)
        self._process_lump(self.delayed_revision_header)
        self.last_revision_passed = rev
        self.delayed_revision_header = None

    def _process_lump(self, lump):
        self._fix_length_fields(lump)
        self._fix_revision_fields(lump)
        self.writer.write_lump(lump)

    def _fix_length_fields(self, lump):
        prop_content_length = self._calculate_prop_content_length(lump)
        if prop_content_length:
            lump.set_header('Prop-content-length', str(prop_content_length))
        elif lump.has_header('Prop-content-length'):
            lump.delete_header('Prop-content-length')

        text_content_length = 0
        if lump.has_header('Text-content-length'):
            text_content_length = int(lump.get_header('Text-content-length'))

        content_length = prop_content_length + text_content_length
        if content_length:
            lump.set_header('Content-length', str(content_length))
        elif lump.has_header('Content-length'):
            lump.delete_header('Content-length')

    def _calculate_prop_content_length(self, lump):
        if not lump.properties:
            return 0
        sum_so_far = 0
        for key in lump.properties.keys():
            key_length = len(key)
            value_length = len(lump.properties[key])
            sum_so_far += (
                3 + len(str(key_length))        # K 4\n
                + key_length + 1                # blub\n
                + 3 + len(str(value_length))    # V 3\n
                + value_length + 1              # xxx\n
            )
        sum_so_far += 10                        # PROPS-END\n
        return sum_so_far

    def _fix_revision_fields(self, lump):
        if lump.has_header('Node-copyfrom-rev'):
            input_revision = int(lump.get_header('Node-copyfrom-rev'))
            mapped_revision = self.revision_mapper.get_output_rev_for_input_rev(input_revision)
            lump.set_header('Node-copyfrom-rev', str(mapped_revision))




class SyntheticDeleter(object):
    def __init__(self, source_repository, lump_builder):
        self.source_repository = source_repository
        self.lump_builder = lump_builder

    def process_revision(self, rev, deleted_paths):
        builder = self.lump_builder
        previous_revision = rev - 1
        builder.revision_header(rev)
        for path in deleted_paths:
            kind = self.source_repository.get_type_of_path(path, previous_revision)
            if kind:
                builder.delete_path(path)




from optparse import OptionParser
from string import join

VERSION = '1.2'

def _parse_command_line(command_line):
    parser = OptionParser("usage: %prog [options] <absolute repository path> <include path> ...\n\nVersion " + VERSION)
    parser.add_option(
        '--include-file', action='append', dest='include_files', default=[],
        help='Read paths to include from given file.',
        metavar='FILE'
    )
    parser.add_option(
        '--exclude', action='append', dest='exclude_paths', default=[],
        help='Exclude given path during filtering.',
        metavar='PATH'
    )
    parser.add_option(
        '--exclude-file', action='append', dest='exclude_files', default=[],
        help='Read paths to exclude from given file.',
        metavar='FILE'
    )
    parser.add_option(
        '--keep-empty-revs', action='store_true', dest='keep_empty_revs', default=False,
        help="Copy revisions even if they have no effect on included paths at all."
    )
    parser.add_option(
        '--start-rev', dest='start_rev', type='int',
        help="Squash the revision history below the given revision. Generate artificial first revision with represents the given input revision.",
        metavar='NUMBER'
    )
    parser.add_option(
        '--no-extra-mkdirs', action='store_false', dest='create_parent_dirs', default=True,
        help="By default extra nodes are injected to create the parent directories of all paths in the include list. Use this option to switch it off."
    )
    parser.add_option(
        '--drop-old-tags-and-branches', action='store_true', dest='drop_old_tags_and_branches', default=False,
        help="Use with the --start-rev option. Automatically exclude data in tags and branches directories referring to data before and up to the start revision."
    )
    parser.add_option(
        '--tag-or-branch-dir', action='append', dest='custom_tags_and_branches_dirs', default=[],
        help='Use with --drop-old-tags-and-branches. Overwrites the list of directory names, which contain tags and branches in your repository. Add one --tag-or-branch-dir option for each name you want - including \'tags\' and \'branches\' if you want to extend the original list.',
        metavar='NAME'
    )
    parser.add_option(
        '--trunk-dir', action='append', dest='custom_trunk_dirs', default=[],
        help='Use with --drop-old-tags-and-branches. Overwrites the list of directory names, which contain the trunk(s) your repository. Add one --tuunk-dir option for each name you want - including \'trunk\' if you want to extend the original list.',
        metavar='NAME'
    )
    parser.add_option(
        '-q', '--quiet', action='store_true', dest='quiet', default=False,
        help="Only log errors and warnings on console."
    )
    parser.add_option(
        '-l', '--log-file', dest='log_file',
        help="Write messages to given log file.",
        metavar='FILE'
    )

    (options, args ) =  parser.parse_args(command_line) 

    if len(args) < 1:
        parser.error('No repository given!')
    source_repository = args[0]
    if not source_repository.startswith('/'):
        parser.error('Please supply ABSOLUTE path to repository!')

    include_paths = [ ]
    for path in args[1:]:
        if path[0] == '/':
            include_paths.append(path[1:])
        else:
            include_paths.append(path)

    return ( options, source_repository, include_paths )
    

def _get_file_as_list(name):
    list = [ ]
    fh = open(name, 'r')
    for line in fh:
        if line[-1:] == "\n":
            line = line[:-1]
        list.append(line)
    return list
        
class Config(object):

    def __init__(self, command_line):
        ( options, source_repository, include_paths ) = _parse_command_line(command_line)

        self.source_repository = source_repository

        for file in options.include_files:
            include_paths += _get_file_as_list(file)
        self.include_paths = include_paths

        exclude_paths = options.exclude_paths
        for file in options.exclude_files:
            exclude_paths += _get_file_as_list(file)
        self.exclude_paths = exclude_paths

        self.keep_empty_revs = options.keep_empty_revs
        self.start_rev = options.start_rev
        self.create_parent_dirs = options.create_parent_dirs
        self.quiet = options.quiet
        self.log_file = options.log_file

        self.drop_old_tags_and_branches = options.drop_old_tags_and_branches

        if options.custom_tags_and_branches_dirs:
            self.tag_and_branch_dict = { }
            for name in options.custom_tags_and_branches_dirs:
                self.tag_and_branch_dict[name] = True
        else:
            self.tag_and_branch_dict = { 'tags': True, 'branches': True }
        self.trunk_dict = { }
        if options.custom_trunk_dirs:
            self.trunk_dict = { }
            for name in options.custom_trunk_dirs:
                self.trunk_dict[name] = True
        else:
            self.trunk_dict = { 'trunk':True }

    def is_path_tag_or_branch(self, path):
        is_just_below_tags_or_branches = False
        is_tag_or_branch = False
        for element in path.split('/'):
            if element == '':
                continue
            if self.trunk_dict.has_key(element):
                return False
            elif self.tag_and_branch_dict.has_key(element):
                is_just_below_tags_or_branches = True
            elif is_just_below_tags_or_branches:
                is_just_below_tags_or_branches = False
                is_tag_or_branch = True
            elif is_tag_or_branch:
                return False
        return is_tag_or_branch



#!/usr/bin/env python

from sys import argv, stdout
from os import putenv
from os.path import basename
from string import join
import logging

from DumpController import DumpController, STRATEGY_DUMP_HEADER, STRATEGY_IGNORE, STRATEGY_SYNTHETIC_DELETES, STRATEGY_DUMP_SCAN, STRATEGY_BOOTSTRAP, DUMP_HEADER_PSEUDO_REV

from DumpHeaderGenerator import DumpHeaderGenerator
from BootsTrapper import BootsTrapper
from DumpFilter import DumpFilter
from SyntheticDeleter import SyntheticDeleter
from RevisionIgnorer import RevisionIgnorer

from LumpBuilder import LumpBuilder
from LumpPostProcessor import LumpPostProcessor

from SvnDumpWriter import SvnDumpWriter
from SvnRepository import SvnRepository

from Config import Config
from InterestingPaths import InterestingPaths
from ParentDirectoryLumpGenerator import ParentDirectoryLumpGenerator
from RevisionMapper import RevisionMapper

console_handler = None

def _setup_interesting_paths(config):
    interesting_paths = InterestingPaths()
    for path in config.include_paths:
        logging.info("Including path '%s'" % (path) )
        interesting_paths.mark_path_as_interesting(path)
    for path in config.exclude_paths:
        logging.info("EXCLUDING path '%s'" % (path) )
        interesting_paths.mark_path_as_boring(path)
    return interesting_paths

def _setup_early_logging():
    global console_handler
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    console_formatter = logging.Formatter('%(levelname)s %(message)s')
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(console_formatter)
    logger.addHandler(console_handler)
    

def _setup_final_logging(config):
    global console_handler
    if config.quiet:
        console_handler.setLevel(logging.WARNING)
    
    if config.log_file:
        log_name = basename(argv[0])
        log_file_formatter = logging.Formatter('%(asctime)s ' + log_name + ' %(levelname)s %(message)s')
        log_file_handler = logging.FileHandler(config.log_file)
        log_file_handler.setLevel(logging.DEBUG)
        log_file_handler.setFormatter(log_file_formatter)
        logger = logging.getLogger()
        logger.addHandler(log_file_handler)

def run():
    _setup_early_logging()
    config = Config(argv[1:])
    _setup_final_logging(config)
    logging.info("Command line: " + join(argv, ' '))
    interesting_paths = _setup_interesting_paths(config)

    # Make sure that we get english error messages
    putenv('LC_MESSAGES', 'C')

    source_repository = SvnRepository(config.source_repository)

    revision_mapper = RevisionMapper(config)
    dump_writer = SvnDumpWriter(stdout)
    with LumpPostProcessor(config, revision_mapper, dump_writer) as lump_post_processor:
        lump_builder = LumpBuilder(config, source_repository, interesting_paths, lump_post_processor)
        parent_directory_lump_generator = ParentDirectoryLumpGenerator(interesting_paths, lump_builder)
        lump_post_processor.parent_directory_lump_generator = parent_directory_lump_generator

        revision_handlers_hash = {
            STRATEGY_DUMP_HEADER:           DumpHeaderGenerator(lump_builder),
            STRATEGY_IGNORE:                RevisionIgnorer(lump_builder),
            STRATEGY_SYNTHETIC_DELETES:     SyntheticDeleter(source_repository, lump_builder),
            STRATEGY_DUMP_SCAN:             DumpFilter(config, source_repository, interesting_paths, lump_builder),
            STRATEGY_BOOTSTRAP:             BootsTrapper(config, source_repository, interesting_paths, lump_builder)
        }

        controller = DumpController(
            config = config,
            repository = source_repository,
            interesting_paths = interesting_paths,
            revision_handlers_by_strategy = revision_handlers_hash
        )

        controller.run()




from os import waitpid
from string import join
from subprocess import Popen, check_call, PIPE, STDOUT
from hashlib import md5
from logging import info
import re

from ContentTin import ContentTin
from CheckedCommandFileHandle import CheckedCommandFileHandle


class RevisionInfo(object):
    def __init__(self, author, date, log_message):
        self.author = author
        self.date = date
        self.log_message = log_message

class TreeHandle(object):
    def __init__(self, file_handle, command):
        self.file_handle = file_handle
        self.command = command
    def __iter__(self):
        return self
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_value, trace):
        self.file_handle = None
        return False
    def next(self):
        line = self.file_handle.readline()
        if not line:
            raise StopIteration()
        if line[-1:] != "\n":
            raise Exception(
                "Line read from '%s' ('%s') has no newline. Truncated?"
                % ( self.command, line )
            )
        return line[:-1]


class SvnRepository(object):

    chunk_size = 1024**2

    def __init__(self, path):
        if path[0] != '/':
            raise Exception("Please supply abosulte path to repository!")
        self.path = path
        self.url = 'file://' + path

    def get_changed_paths_by_change_type_for_revision(self, rev):
        changes = { }
        with CheckedCommandFileHandle([ 'svnlook', 'changed', '-r', str(rev), self.path ]) as fh:
            for line in fh:
                change_type = line[:3].strip()
                path = line[4:-1]
                if line[-1:] != "\n":
                    raise Exception(
                        "Line read from 'svnlook changed -r %d %s' has no newline. Truncated?"
                        % ( rev, self.path )
                    )
                if changes.has_key(change_type):
                    changes[change_type].append(path)
                else:
                    changes[change_type] = [ path ]

        return changes

    def get_tin_for_file(self, path, rev):
        # It is ugly to read the file twice. But we don't want to keep the whole file in
        # memory. And:  We just need to calculate the MD5 checksum BEFORE we copy it to the output.
        md5_calculator = md5()
        size = 0
        with CheckedCommandFileHandle([ 'svnlook', 'cat', '-r', str(rev), self.path, path]) as fh:
            while True:
                chunk = fh.read(self.chunk_size)
                if chunk == '':
                    break
                md5_calculator.update(chunk)
                size += len(chunk)
        md5sum = md5_calculator.hexdigest()
        fh = CheckedCommandFileHandle([ 'svnlook', 'cat', '-r', str(rev), self.path, path])
        return ContentTin(fh, size, md5sum)

    def get_properties_of_path(self, path, rev):
        if path[-1:] == '/':
            path = path[:-1]
        property_list = [ ]
        rev_string = str(rev)
        with CheckedCommandFileHandle([ 'svnlook', 'pl', '-r', rev_string, self.path, path]) as fh:
            for line in fh:
                if line[-1:] != "\n":
                    raise Exception(
                        "Line read from 'svnlook pl -r %d %s %s' has no newline. Truncated?"
                        % ( rev, self.path, path )
                    )
                if line[:2] == '  ':
                    property_list.append(line[2:-1])

        properties = { }
        for property_name in property_list:
            with CheckedCommandFileHandle([ 'svnlook', 'pg', '-r', rev_string, self.path, property_name, path]) as fh:
                property_value = fh.read()
            properties[property_name] = property_value

        return properties

    def get_type_of_path(self, path, rev):
        if path[-1:] == '/':
            path = path[:-1]
        args = [ 'svnlook', 'tree', '--full-paths', '-r', str(rev), self.path, path ]
        process = Popen(args, stdout=PIPE, stderr=PIPE)
        output = process.stdout.read()
        error = process.stderr.read()
        process.stdout.close()
        process.stderr.close()
        status = waitpid(process.pid, 0)[1]
        if status:
            if re.match('svnlook: (?:.*)File not found:', error):
                return None
            else:
                 raise Exception(
                    "Command '%s' exited unexpected with status %d and error message:\n%s"
                    % ( join(args), status, error )
                )
        if output.startswith(path + "/\n"):
            return 'dir'
        else:
            return 'file'

    def get_dump_file_handle_for_revision(self, rev):
        return CheckedCommandFileHandle(
            [ 'svnadmin', 'dump', '--incremental', '-r', str(rev), self.path ],
            [   # SVN 1.6
                '^\* Dumped revision \d+\.$',
                '^WARNING: Referencing data in revision \d+, which is older than the oldest$',
                '^WARNING: dumped revision \(\d+\)\.  Loading this dump into an empty repository$',
                '^WARNING: will fail\.$',
                # SVN 1.7
                'WARNING 0x0000: Referencing data in revision \d+, which is older than the oldest dumped revision \(r\d+\)',
                'WARNING 0x0000: The range of revisions dumped contained references to copy sources outside that range\.',
                'WARNING 0x0001: Mergeinfo referencing revision\(s\) prior to the oldest dumped revision \(r\d+\)\. Loading this dump may result in invalid mergeinfo\.',
                'WARNING 0x0001: The range of revisions dumped contained mergeinfo which reference revisions outside that range\.',
            ]
        )

    def get_tree_handle_for_path(self, path, rev):
        args = [ 'svnlook', 'tree', '--full-paths', '-r', str(rev), self.path, path]
        list = [ ]
        fh = CheckedCommandFileHandle(args)
        command = join(args, ' ')
        return TreeHandle(fh, command)

    def get_revision_info(self, rev):
        properties = { }
        for property_name in [ 'svn:author', 'svn:date', 'svn:log' ]:
            with CheckedCommandFileHandle([ 'svnlook', 'pg', '--revprop', '-r', str(rev), self.path, property_name]) as fh:
                properties[property_name] = fh.read()
        return RevisionInfo(
            author = properties['svn:author'],
            date = properties['svn:date'],
            log_message = properties['svn:log'],
        )

    def get_uuid(self):
        with CheckedCommandFileHandle([ 'svnlook', 'uuid', self.path ]) as fh:
            line = fh.read()
        return line[:-1]

    def get_youngest_revision(self):
        with CheckedCommandFileHandle([ 'svnlook', 'youngest', self.path ]) as fh:
            line = fh.read()
        return int(line[:-1])




INHERITED = "INHERITED"
BORING = "BORING"
INTERESTING = "INTERESTING"

class _PathNode(object):
    def __init__(self, path, node_type=INHERITED):
        self.children = { }
        self.node_type = node_type
        self.path = path

    def add_child(self, name):
        if self.path:
            path = self.path + '/' + name
        else:
            path = name
        child = _PathNode(path)
        self.children[name] = child
        return child

    def __str__(self):
        return "NODE(%s) -> %s" % ( self.node_type, self.children )

    def __repr__(self):
        return str(self)

def split_path(path):
    splitted_path = path.split("/") 
    return [x for x in splitted_path if x != ""]

class InterestingPaths(object):

    def __init__(self):
        self.root_node = _PathNode(None)

    def mark_path_as_interesting(self, path):
        self._mark_path_as_type(path, INTERESTING)

    def mark_path_as_boring(self, path):
        self._mark_path_as_type(path, BORING)

    def _mark_path_as_type(self, path, node_type):
        path_elements = split_path(path)
        current_node = self.root_node
        for element in path_elements:
            if element in current_node.children:
                current_node = current_node.children[element]
            else:
                current_node = current_node.add_child(element)
        if current_node.node_type != INHERITED or current_node.children:
            raise Exception("path inconsistent: " + path)
        current_node.node_type = node_type

    def is_interesting(self, path):
        ( node, node_type ) = self._get_node_and_type_of_path(path)
        return node_type == INTERESTING

    def _get_node_and_type_of_path(self, path):
        path_elements = split_path(path)
        current_node = self.root_node
        current_type = current_node.node_type
        for element in path_elements:
            if element in current_node.children:
                current_node = current_node.children[element]
                if current_node.node_type != INHERITED:
                    current_type = current_node.node_type
            else:
                current_node = None
                break
        return ( current_node, current_type )

    def get_interesting_sub_directories(self, path):
        ( node, node_type ) = self._get_node_and_type_of_path(path)
        if node_type == INTERESTING:
            return [ path ]
        if node is None:
            return [ ]
        return self._get_interesting_sub_directories_for_node(node)

    def _get_interesting_sub_directories_for_node(self, node):
        if node.node_type == INTERESTING:
            return [ node.path ]
        dirs_found = [ ]
        for key in node.children.keys():
            child = node.children[key]
            dirs_found += self._get_interesting_sub_directories_for_node(child)
        return dirs_found




from logging import info
from SvnDumpReader import SvnDumpReader
from LumpBuilder import LumpBuilder

class UnsupportedDumpVersionException(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class DumpFilter(object):
    def __init__(self, config, source_repository, interesting_paths, lump_builder):
        self.config = config
        self.source_repository = source_repository
        self.interesting_paths = interesting_paths
        self.dump_reader = None
        self.revision_number = None
        self.lump_builder = lump_builder

    def process_revision(self, revision, aux_data):
        assert aux_data is None
        input_fh = self.source_repository.get_dump_file_handle_for_revision(revision)
        self.dump_reader = SvnDumpReader(input_fh)
        self._process_header_lumps()

        while True:
            lump = self.dump_reader.read_lump()
            if lump is None:
                break;
            self._process_lump(lump) 

        self.dump_reader = None

    def _process_header_lumps(self):
        lump = self.dump_reader.read_lump()
        while lump and not lump.has_header('Revision-number'):
            self._validate_dump_header_lump(lump)
            lump = self.dump_reader.read_lump()
        if not lump:
            raise Exception("Failed to parse dump of revision %d: No revision header found!")
        self.revision_number = int(lump.get_header('Revision-number'))
        self.lump_builder.pass_lump(lump)

    def _validate_dump_header_lump(self, lump):
        if lump.has_header('SVN-fs-dump-format-version'):
            dump_format_version = lump.get_header('SVN-fs-dump-format-version')
            if dump_format_version != '2':
                raise UnsupportedDumpVersionException(
                    'Detected unsupported SVN version - SVN-fs-dump-format-version was %s (wanted: 2)'
                    % ( dump_format_version )
                )

    def _process_lump(self, lump):
        action = lump.get_header('Node-action')
        if action == 'add' or action == 'replace':
            self._process_add_or_replace_lump(lump, action)
        elif action == 'delete':
            self._process_delete_lump(lump)
        elif action == 'change':
            self._process_change_lump(lump)
        else:
            raise Exception("Unknown Node-action '%s'!" % ( action ) )

    def _process_add_or_replace_lump(self, lump, action):
        path = lump.get_header('Node-path')
        if self.interesting_paths.is_interesting(path):
            self._process_add_or_replace_lump_of_interesting_path(lump, action)
            return

        if lump.get_header('Node-kind') != 'dir':
            return
        if not lump.has_header('Node-copyfrom-path'):
            return

        from_path = lump.get_header('Node-copyfrom-path')
        from_rev = int(lump.get_header('Node-copyfrom-rev'))
        self._add_or_replace_interesting_sub_directories_of_path(path, action, from_path, from_rev)

    def _process_add_or_replace_lump_of_interesting_path(self, lump, action):
        path = lump.get_header('Node-path')
        if not lump.has_header('Node-copyfrom-path'):
            self.lump_builder.pass_lump(lump)
            return

        from_path = lump.get_header('Node-copyfrom-path')
        from_rev = int(lump.get_header('Node-copyfrom-rev'))

        if self._is_internal_copy(from_path, from_rev):
            self.lump_builder.pass_lump(lump)
        else:
            node_kind = lump.get_header('Node-kind')
            self.lump_builder.get_recursively_from_source(node_kind, path, action, from_path, from_rev)
            if lump.content:
                self.lump_builder.change_lump_from_add_or_replace_lump(lump)

    def _is_internal_copy(self, from_path, from_rev):
        if self.config.start_rev and from_rev <  self.config.start_rev:
            return False
        return self.interesting_paths.is_interesting(from_path)

    def _add_or_replace_interesting_sub_directories_of_path(self, path, action, from_path, from_rev):
        if path[-1:] == '/':
            path = path[:-1]
        paths_to_check = self.interesting_paths.get_interesting_sub_directories(path)
        for sub_path in paths_to_check:
            from_sub_path = from_path + sub_path[len(path):]
            node_kind = self.source_repository.get_type_of_path(from_sub_path, from_rev)
            if not node_kind:
                continue
            if self._is_internal_copy(from_sub_path, from_rev):
                self.lump_builder.get_path_from_target(node_kind, sub_path, action, from_sub_path, from_rev)
            else:
                self.lump_builder.get_recursively_from_source(node_kind, sub_path, action, from_sub_path, from_rev)

    def _process_delete_lump(self, lump):
        path = lump.get_header('Node-path')
        if self.interesting_paths.is_interesting(path):
            self.lump_builder.pass_lump(lump)
            return

        paths_to_check = self.interesting_paths.get_interesting_sub_directories(path)
        if not paths_to_check:
            return
        previous_revision = self.revision_number - 1
        for path_to_check in paths_to_check:
            does_exist = self.source_repository.get_type_of_path(path_to_check, previous_revision) is not None
            if does_exist:
                self.lump_builder.delete_path(path_to_check)

    def _process_change_lump(self, lump):
        path = lump.get_header('Node-path')
        if self.interesting_paths.is_interesting(path):
            self.lump_builder.pass_lump(lump)






class BootsTrapper(object):
    def __init__(self, config, source_repository, interesting_paths, lump_builder):
        self.config = config
        self.source_repository = source_repository
        self.interesting_paths =interesting_paths
        self.lump_builder = lump_builder

    def process_revision(self, revision, aux_data):
        assert aux_data is None
        builder = self.lump_builder
        builder.revision_header(revision, 'svnfiltereddump boots trap revision')
        paths_of_interest = sorted(self.interesting_paths.get_interesting_sub_directories(''))
        for path in paths_of_interest:
            kind = self.source_repository.get_type_of_path(path, revision)
            if kind:
                builder.get_recursively_from_source(kind, path, 'add', path, revision)    




from logging import info

STRATEGY_DUMP_HEADER = 'DUMP_HEADER'
STRATEGY_IGNORE = 'IGNORE'
STRATEGY_SYNTHETIC_DELETES = 'SYNTHETIC_DELETES'
STRATEGY_DUMP_SCAN = 'DUMP_SCAN'
STRATEGY_BOOTSTRAP = 'BOOTSTRAP'

DUMP_HEADER_PSEUDO_REV = -1

class DumpController(object):
    def __init__(self, config, repository, interesting_paths, revision_handlers_by_strategy):
        self.config = config
        self.repository = repository
        self.interesting_paths = interesting_paths
        self.revision_handlers_by_strategy = revision_handlers_by_strategy

    def _get_strategy_and_aux_data_for_revision(self, rev):
        if rev == DUMP_HEADER_PSEUDO_REV:
            return ( STRATEGY_DUMP_HEADER, None )
        if rev == self.config.start_rev:
            return ( STRATEGY_BOOTSTRAP, None )

        changes_by_type = self.repository.get_changed_paths_by_change_type_for_revision(rev)

        for change_type in changes_by_type.keys():
            if change_type == 'D':
                continue
            for path in changes_by_type[change_type]:
                if self.interesting_paths.is_interesting(path):
                    return ( STRATEGY_DUMP_SCAN, None )
                if change_type == 'A':
                    if self.interesting_paths.get_interesting_sub_directories(path):
                        return ( STRATEGY_DUMP_SCAN, None )

        if not changes_by_type.has_key('D'):
            return ( STRATEGY_IGNORE, None )

        delete_paths = [ ]
        for path in changes_by_type['D']:
            delete_paths += self.interesting_paths.get_interesting_sub_directories(path)
        if delete_paths:
            return ( STRATEGY_SYNTHETIC_DELETES, delete_paths )
       
        return ( STRATEGY_IGNORE, None )

    def run(self):
        header_handler = self.revision_handlers_by_strategy[STRATEGY_DUMP_HEADER]
        header_handler.process_revision(DUMP_HEADER_PSEUDO_REV, None)

        first_revision = 1
        if self.config.start_rev:
            first_revision = self.config.start_rev
        last_revision = self.repository.get_youngest_revision()

        for revision in xrange(first_revision, last_revision+1):
            ( strategy, aux_data ) = self._get_strategy_and_aux_data_for_revision(revision)
            info("Processing input revsion %d using strategy %s ..." % ( revision, strategy ) )
            handler = self.revision_handlers_by_strategy[strategy]
            handler.process_revision(revision, aux_data) 



#!/usr/bin/env python

from distutils.core import setup

setup(
    name='svnfiltereddump',
    version='1.2',
    description='Extracts parts from Subversion repositories',
    author='Harald Wilhelmi',
    author_email='harald.wilhelmi@tngtech.com',
    url='https://github.com/tng/svnfiltereddump',
    packages=['svnfiltereddump'],
    package_dir = { '': 'src' },
    long_description="""\
        The svnfiltereddump tool allows to extract parts of Subversion
        repositories. It is a bit like svndumpfilter or svndumpfilter2.
        But it has less limitations and has the ability to drop drop
        old revisions.
    """,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Version Control",
    ],
    keywords='subversion',
    license='GPL',
    scripts = [ 'src/bin/svnfiltereddump' ]
)




from re import match
from SvnLump import SvnLump
from ContentTin import ContentTin
from string import join

class SvnDumpReader(object):

    def __init__(self, file_handle):
        self.file_handle = file_handle
        self.current_lump = None

    def read_lump(self):
        self._finish_last_lump()
        self.current_lump = SvnLump()
        self._read_headers()
        if not self.current_lump.get_header_keys():
            return None
        self._read_properties()
        self._create_content_tin()
        return self.current_lump

    def _finish_last_lump(self):
        # Ensure that ContentTin of last lump gets depleted
        if self.current_lump and self.current_lump.content:
            self.current_lump.content.discard()

    def _read_headers(self):
        fh = self.file_handle
        header_count = 0

        while True:
            line = fh.readline()
            if not line:
                if header_count > 0:
                    raise Exception("Found End of file before end of headers!")
                return
            if line == "\n":
                if header_count > 0:
                    return
            else:
                m = match("(.+): (.*)$", line)
                if not m:
                    raise Exception("Found broken header line:\n" + line)
                self.current_lump.set_header(m.group(1), m.group(2))
                header_count += 1

    def _read_properties(self):
        lump = self.current_lump
        if not lump.has_header('Prop-content-length'):
            return
        expected_length = int(lump.get_header('Prop-content-length'))
        actual_length = 0

        while True:
            ( key, key_length ) = self._read_prop_key_and_length()
            actual_length += key_length
            if key is None:
                break;
            ( value, value_length ) = self._read_prop_value_and_length()
            actual_length += value_length
            lump.properties[key] = value
        if actual_length != expected_length:
            raise Exception(
                "PROPS section should be %d bytes long but was %d!"
                % ( expected_length, actual_length )
            )

    def _read_prop_key_and_length(self):
        fh = self.file_handle
        line = fh.readline()
        length = len(line)
        if line == 'PROPS-END\n':
            return ( None, length )
        if not line.startswith('K '):
            raise Exception("Needed key size line (K <number>) but got:\n" + line)
        size = int(line[2:-1])
        ( key, key_length ) = self._read_field_of_given_size_and_length(size)
        length += key_length
        return ( key, length )
        
    def _read_prop_value_and_length(self):
        fh = self.file_handle
        line = fh.readline()
        length = len(line)
        if not line.startswith('V '):
            raise Exception("Needed value size line (V <number>) but got:\n" + line)
        size = int(line[2:-1])
        ( value, value_length ) = self._read_field_of_given_size_and_length(size)
        length += value_length
        return ( value, length )

    def _read_field_of_given_size_and_length(self, size):
        lines = [ ]
        length = 0
        while True:
            line = self.file_handle.readline()
            if not line:
                raise Exception("Reached end-of-file while reading field!")
            lines.append(line)
            length += len(line)
            if length == size+1:
                return ( join(lines, '')[0:-1], length )
            elif length > size+1:
                raise Exception("Field length did not match expected size!")

    def _create_content_tin(self):
        lump = self.current_lump
        if not lump.has_header('Text-content-length'):
            return
        size = int(lump.get_header('Text-content-length'))
        md5sum = None
        if lump.has_header('Text-content-md5'):
            md5sum = lump.get_header('Text-content-md5')
        lump.content = ContentTin(self.file_handle, size, md5sum)



#
# This class is a wrapper around subprocess.Popen
#
# It behaves like a the stdout pipe generated by Popen.
# However it makes a serious efford to do this:
# 1) Forward STDERR of generated process to our STDERR
# 2) Don't get stuck - no matter how much output we get
#    on STDOUT or STDERR
# 3) Raise an Exception if the exit code was not 0
# 4) Raise an Exception if there was data on STDERR
#

from fcntl import fcntl, F_GETFL, F_SETFL
from os import waitpid, O_NONBLOCK
from sys import stderr
from subprocess import Popen, check_call, PIPE, STDOUT
from string import join
from errno import EAGAIN
import re

class CheckedCommandFileHandle(object):

    def __init__(self, args, ignore_patterns=[], error_fh=stderr):
        self.status_ok = True
        self.args = args
        self.ignore_patterns = ignore_patterns
        self.error_fh = error_fh
        self.process = Popen(args, stdout=PIPE, stderr=PIPE)
        self._make_stderr_non_blocking()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        if exc_value:
            self._empty_stderr()
        else:
            self.close()
        return False

    def __iter__(self):
        return self.process.stdout

    def _make_stderr_non_blocking(self):
        file_no = self.process.stderr.fileno()
        bits = O_NONBLOCK | fcntl(file_no, F_GETFL)
        fcntl(file_no, F_SETFL, bits)

    def next(self):
        self._empty_stderr()
        return self.process.stdout.next()

    def readline(self, size=-1):
        self._empty_stderr()
        return self.process.stdout.readline(size)

    def read(self, size=-1):
        self._empty_stderr()
        return self.process.stdout.read(size)

    def _empty_stderr(self):
        try:
            error = self.process.stderr.read()
        except IOError as (errno, strerror):
            if errno == EAGAIN:
                return
            raise
        self._validate_error_output(error)

    def _validate_error_output(self, error):
        if not error:
            return
        if error[-1:] == "\n":
            error = error[:-1]
        error_lines = error.split("\n")
        line_ok = False
        for line in error_lines:
            line_ok = False
            for pattern in self.ignore_patterns:
                if re.search(pattern, line):
                    line_ok = True
                    break
            if not line_ok:
                self.error_fh.write("FAILED LINE: '"+line+"'\n")
                cmd = join(self.args, ' ')
                self.error_fh.write("Output of command '%s' on STDERR:\n%s\n" % ( cmd, error ) )
                raise Exception("Command '%s' wrote to STDERR (see above)!" % ( cmd ))

    def close(self):
        p = self.process
        self._empty_stderr()
        p.stdout.close()
        self._empty_stderr()
        p.stderr.close()
        status = waitpid(p.pid, 0)[1]
        if status:
            raise Exception(
                "Command '%s' exited with status %d"
                % ( join(self.args), status )
            )





from os.path import normpath
from copy import copy
from logging import info

from SvnLump import SvnLump
from SvnRepository import SvnRepository


class LumpBuilder(object):
    def __init__(self, config, source_repository, interesting_paths, dump_writer):
        self.config = config
        self.source_repository = source_repository
        self.interesting_paths = interesting_paths
        self.dump_writer = dump_writer

    def delete_path(self, path):
        lump = SvnLump()
        lump.set_header('Node-path', normpath(path))
        lump.set_header('Node-action', 'delete')
        self.dump_writer.write_lump(lump)

    def mkdir(self, path):
        lump = SvnLump()
        lump.set_header('Node-path', normpath(path))
        lump.set_header('Node-kind', 'dir')
        lump.set_header('Node-action', 'add')
        self.dump_writer.write_lump(lump)

    def get_node_from_source(self, kind, path, action, from_path, from_rev):
        assert kind == 'file' or kind == 'dir'
        assert action == 'add' or action =='replace'

        if self._is_obsolete_tag_or_branch_copy(path, from_rev):
            self._handle_obsolete_tag_or_branch(path)
            return

        path = normpath(path)
        repo = self.source_repository
        lump = SvnLump()

        lump.set_header('Node-path', path)
        lump.set_header('Node-kind', kind)
        lump.set_header('Node-action', action)

        if kind == 'file':
            tin = repo.get_tin_for_file(from_path, from_rev)
            lump.content = tin
            lump.set_header('Text-content-length', str(tin.size))
            lump.set_header('Text-content-md5', tin.md5sum)

        lump.properties = repo.get_properties_of_path(from_path, from_rev)
        self.dump_writer.write_lump(lump)

    def _is_obsolete_tag_or_branch_copy(self, path, from_rev):
        if not self.config.start_rev:
            return False
        if not self.config.drop_old_tags_and_branches:
            return False
        return self.config.is_path_tag_or_branch(path)

    def _handle_obsolete_tag_or_branch(self, path):
        info('Excluding obsolete tag/branch: ' + path)
        self.interesting_paths.mark_path_as_boring(path)

    def get_path_from_target(self, kind, path, action, from_path, from_rev):
        assert kind == 'file' or kind =='dir'
        assert action == 'add' or action =='replace'

        lump = SvnLump()
        lump.set_header('Node-path', path)
        lump.set_header('Node-kind', kind)
        lump.set_header('Node-action', action)
        lump.set_header('Node-copyfrom-path', from_path)
        lump.set_header('Node-copyfrom-rev', str(from_rev))
        self.dump_writer.write_lump(lump)

    def change_lump_from_add_or_replace_lump(self, sample_lump):
        lump = copy(sample_lump)
        lump.set_header('Node-action', 'change')
        for header_name in [
            'Node-kind', 'Node-copyfrom-path', 'Node-copyfrom-rev',
            'Text-copy-source-md5', 'Text-copy-source-sha1'
        ]:
            if lump.has_header(header_name):
                lump.delete_header(header_name)
        self.dump_writer.write_lump(lump)

    def revision_header(self, rev, log_message=None):
        lump = SvnLump()
        lump.set_header('Revision-number', str(rev))
        rev_info = self.source_repository.get_revision_info(rev)
        if log_message:
            new_log_message = log_message
        else:
            new_log_message = rev_info.log_message
        lump.properties = {
            'svn:author': rev_info.author,
            'svn:date': rev_info.date,
            'svn:log': new_log_message
        }
        self.dump_writer.write_lump(lump)

    def dump_header_lumps(self):
        lump = SvnLump()
        lump.set_header('SVN-fs-dump-format-version', '2')
        self.dump_writer.write_lump(lump)

        lump = SvnLump()
        lump.set_header('UUID', str(self.source_repository.get_uuid()))
        self.dump_writer.write_lump(lump)

    def pass_lump(self, lump):
        self.dump_writer.write_lump(lump)

    def get_recursively_from_source(self, kind, path, action, from_path, from_rev):
        assert kind == 'file' or kind =='dir'
        assert action == 'add' or action =='replace'

        if self._is_obsolete_tag_or_branch_copy(path, from_rev):
            self._handle_obsolete_tag_or_branch(path)
            return
        if kind == 'file':
            self.get_node_from_source(kind, path, action, from_path, from_rev)
        else:
            if action == 'replace':
                self.delete_path(path)
            self._add_tree_from_source(path, from_path, from_rev)

    def _add_tree_from_source(self, path, from_path, from_rev):
        if path[-1:] != '/' and len(path)>0:
            path += '/'
        if from_path[-1:] != '/':
            from_path += '/'
        with self.source_repository.get_tree_handle_for_path(from_path, from_rev) as tree_handle:
            for from_sub_path in tree_handle:
                if from_sub_path[-1:] == '/':
                    kind = 'dir'
                    from_sub_path = from_sub_path[:-1]
                else:
                    kind = 'file'
                sub_path = path + from_sub_path[len(from_path):]
                if sub_path == '':
                    continue
                if self.interesting_paths.is_interesting(sub_path):
                    self.get_node_from_source(kind, sub_path, 'add', from_sub_path, from_rev)




from logging import info

class SvnDumpWriter(object):
    def __init__(self, file_handle):
        self.file_handle = file_handle

    def write_lump(self, lump):
        self._write_headers(lump)
        props_written = self._write_properties(lump)
        content_written = self._write_content(lump)
        if props_written or content_written:
            self.file_handle.write("\n") 

    def _write_headers(self, lump):
        fh = self.file_handle
        for header_name in lump.get_header_keys():
            header_value = lump.get_header(header_name)
            fh.write(
                "%s: %s\n"
                % ( header_name, header_value )
            )
        fh.write("\n") 

    def _write_properties(self, lump):
        if not lump.properties:
            return False
        fh = self.file_handle
        for key in sorted(lump.properties.keys()):  # Sorted for easier testing
            value = lump.properties[key]
            fh.write(
                "K %d\n%s\nV %d\n%s\n"
                % ( len(key), key, len(value), value )
            )
        fh.write("PROPS-END\n")
        return True

    def _write_content(self, lump):
        tin = lump.content
        if tin:
            tin.empty_to(self.file_handle)
            self.file_handle.write("\n") 
            return True
        else:
            return False
